
#define COMMAND_HELLO                 0x11
#define COMMAND_PIN_MODE              0x22
#define COMMAND_DIGITAL_OUT           0x33
#define COMMAND_DIGITAL_IN            0x44
#define COMMAND_ANALOG_IN             0x55

#define COMMAND_ERROR                 0xff

#define ERROR_UNKNOWN_COMMAND         0x01
#define ERROR_BUFFER_TOO_SMALL        0x02
#define ERROR_WRONG_DATA_LEN          0x03


uint8_t pinStatus[] = {
  0, 0, 0, 0, 0,
  0, 0, 0, 0, 0,
  0, 0, 0, 0, 0,
  0, 0, 0, 0, 0};

uint32_t ErrorCounter = 0;

// UART Input Data structures
enum InputState_t {STATE_STARTFRAME, STATE_LEN, STATE_COMMAND, STATE_TRASH};
struct SerialInput_t {
  uint8_t len;
  uint8_t command;
  InputState_t inputstate;
} SerialInput;

void ClearSerialInput() {
  while (SerialInput.len > 0) {
    Serial.read();
    SerialInput.len--;
  }
  SerialInput.command = 0;
  SerialInput.inputstate = STATE_STARTFRAME;
}

void SendErrorCode(uint8_t code) {
  Serial.write(0xde); // frame start
  Serial.write(0xad); // frame start
  Serial.write(0x02); // length
  Serial.write(COMMAND_ERROR);
  Serial.write(code);
}

void SendChunk(uint8_t command, uint8_t *chunk, uint16_t chunklen = 0) {
  Serial.write(0xde); // frame start
  Serial.write(0xad); // frame start
  Serial.write((chunklen+1)&0xff);
  Serial.write(command);
    
  if (chunklen>0)
    Serial.write(chunk, chunklen);
}

void setup() {
  // Init UART
  Serial.begin(115200);
  SerialInput.len = 0;
  ClearSerialInput();
}

void loop() {
  switch (SerialInput.inputstate) {
    case STATE_STARTFRAME:
        if (Serial.available() > 1) {
          uint8_t first = Serial.read();
          uint8_t second = Serial.peek();
          
          if ( (first == 0xde) && (second == 0xad) ) {
            Serial.read();
            SerialInput.inputstate = STATE_LEN;
          }
        }
        break;
        
    case STATE_LEN: // next serial input is chunk length (1 Byte)
        if (Serial.available() > 0) {
          SerialInput.len = Serial.read();
          if (SerialInput.len==0) {
            SendErrorCode(ERROR_UNKNOWN_COMMAND);
            ClearSerialInput();
          } else if (SerialInput.len>64)
            SerialInput.inputstate = STATE_TRASH;
          else
            SerialInput.inputstate = STATE_COMMAND;
        }
        break;
        
    case STATE_COMMAND: // next serial input is chunk data (length from STATE_LEN step)
        // execute command only, when all data is received
        if (Serial.available() >= SerialInput.len) {
          uint8_t command = Serial.read();
          SerialInput.len--;
          
          switch (command) {
            
            case COMMAND_HELLO:
                if (SerialInput.len>0)
                  SendErrorCode(ERROR_WRONG_DATA_LEN);
                else {
                  uint8_t res = 1;
                  SendChunk(COMMAND_HELLO, &res, 1);
                }
                break;
            
            case COMMAND_PIN_MODE:
                if (SerialInput.len!=5)
                  SendErrorCode(ERROR_WRONG_DATA_LEN);
                else {
                  uint8_t pin = 0;
                  for (uint8_t b=0; b<5; b++) {
                    uint8_t databyte = Serial.read();
                    for (uint8_t i=0; i<4; i++) {
                      if ((pin>1) & (pin<NUM_DIGITAL_PINS)) { // don't change serial port settings & non available pins
                        pinStatus[pin] = (databyte>>6) & 0b11;
                        
                        if (pinStatus[pin] == 0b11)
                          pinMode(pin,  INPUT_PULLUP);
                        else if (pinStatus[pin] != 0b00)
                          pinMode(pin, INPUT);
                        else
                          pinMode(pin, OUTPUT);
                         
                      }
                      databyte <<=2;
                      pin++;
                    }
                  }
                  
                  uint8_t res = 1;
                  SendChunk(COMMAND_PIN_MODE, &res, 1);
                  SerialInput.len = 0;
                }
                break;
                
            case COMMAND_DIGITAL_OUT:
                if (SerialInput.len!=3)
                  SendErrorCode(ERROR_WRONG_DATA_LEN);
                else {
                  uint8_t pin = 0;
                  for (uint8_t b=0; b<3; b++) {
                    uint8_t databyte = Serial.read();
                    for (uint8_t i=0; i<8; i++) {
                      if ((pin>1) & (pin<NUM_DIGITAL_PINS)) { // don't change serial port settings & non available pins
                        
                        if (pinStatus[pin]==0) // operate on output pins only
                          digitalWrite(pin, (databyte&0x80) ? HIGH : LOW);
                      }
                      databyte <<=1;
                      pin++;
                    }
                  }
                  
                  uint8_t res = 1;
                  SendChunk(COMMAND_DIGITAL_OUT, &res, 1);
                  SerialInput.len = 0;
                }
                break;     
                
            case COMMAND_DIGITAL_IN:
                if (SerialInput.len!=0)
                  SendErrorCode(ERROR_WRONG_DATA_LEN);
                else {
                  uint8_t pin = 0;
                  uint8_t chk[3];
                  for (uint8_t b=0; b<3; b++) {
                    chk[b] = 0;
                    for (uint8_t i=0; i<8; i++) {
                      chk[b] <<= 1;
                      if ((pin>1) & (pin<NUM_DIGITAL_PINS)) {
                        if ( (pinStatus[pin]&1) && (digitalRead(pin)))
                          chk[b] |= 1;
                      }
                      
                      pin++;
                    }                    
                  }
                  
                  SendChunk(COMMAND_DIGITAL_IN, &chk[0], 3);
                }
                break;              
                
            case COMMAND_ANALOG_IN:
                if (SerialInput.len!=0)
                  SendErrorCode(ERROR_WRONG_DATA_LEN);
                else {
                  uint8_t chk[6];
                  for (uint8_t pin=0; pin<6; pin++) {
                    chk[pin] = 0;
                    if (pinStatus[pin + A0] == 0b10) {
                      uint16_t val = analogRead(pin + A0) >> 2;
                      chk[pin] = (uint8_t)val;
                    }                      
                  }
                  
                  SendChunk(COMMAND_ANALOG_IN, &chk[0], 6);
                }
                break;        
            default:
              SendErrorCode(ERROR_UNKNOWN_COMMAND);
              
          }

          ClearSerialInput();
        }
        break;  
        
    case STATE_TRASH: // next serial input is trash
        if (Serial.available() > 0) {
          Serial.read();
          SerialInput.len--;
          if (SerialInput.len == 0) {
            SendErrorCode(ERROR_BUFFER_TOO_SMALL);
            ClearSerialInput();
          }
        }
        break;
  }
  
  // Error Timeout
  if (SerialInput.inputstate != STATE_STARTFRAME) {
    ErrorCounter++;
    if (ErrorCounter == 0) {
      SerialInput.len = 0;
      ClearSerialInput();
    }
  } else {
    ErrorCounter = 0;
  }
}
